#!/usr/bin/env python

# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# Merchantability Or Fitness FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import h5py
import sys
import json
import numpy as np
from scipy.optimize import curve_fit
from collections import Counter
import feature_extractor as fx
import logging

from allensdk.core.nwb_data_set import NwbDataSet


def extract_cell_features(nwb_file, long_square_sweeps, short_square_sweeps, ramp_sweeps):
    '''
    Extract features of long square, short square, and ramp sweeps used for cell-wide characterization.
    
    Parameters
    ----------

    nwb_file: string
        File name of NWB file.

    long_square_sweeps: list
        List of sweep numbers of long square sweeps in an NWB file.
        
    short_square_sweeps: list
        List of sweep numbers of short square sweeps in an NWB file.

    ramp_sweeps: list
        List of sweep numbers of ramp sweeps in an NWB file.

    Returns
    -------

    dict: dictionary of properties grouped by sweep type

    
    '''

    long_square_features = analyze_long_squares(long_square_sweeps, nwb_file)

    if ( ( "hero_sweep_num" in long_square_features ) and ( "input_resistance" in long_square_features ) ):
        return {
            'long_squares': long_square_features,
            'ramps': analyze_ramps(ramp_sweeps, nwb_file),
            'short_squares': analyze_short_squares(short_square_sweeps, nwb_file)
            }

    else:
        return None


def extract_sweep_features(nwb_file, sweep_numbers):
    '''
    Run feature extraction on a per sweep basis for a set of sweeps.  This does not
    feature aggregation for the set of the sweeps.

    Parameters
    ----------

    nwb_file: string
        File name of an NWB file

    sweep_numbers: list
        List of sweeps numbers.
    
    '''

    features = fx.EphysFeatureExtractor()
    all_sweep_features = {}

    for sweep_number in sweep_numbers:
        sweep_features = extract_single_sweep_features(features, nwb_file, sweep_number)
        all_sweep_features[sweep_number] = sweep_features

    return all_sweep_features

def extract_single_sweep_features(features, nwb_file, sweep_number):
    '''
    Run feature extraction on a single sweep.  

    Parameters
    ----------
    
    features: EphysFeatureExtractor instance
    
    nwb_file: string
        File name of an NWB file

    sweep_numbers: int
        Sweep number in the NWB file
       
    '''

    nwb = NwbDataSet(nwb_file)
    data = nwb.get_sweep(sweep_number)

    v = data['response']
    curr = data['stimulus']

    idx0 = data['index_range'][0]
    idx1 = data['index_range'][1]

    if idx0 >= idx1:
        logging.warning("Sweep %s stop index precedes start index, skipping spike identification" % sweep_number)
        return 

    hz = data['sampling_rate']
    dt = 1.0 / hz
    t = np.arange(0, len(v)) * dt
    
    features.process_instance(sweep_number, v*1e3, curr*1e12, t, dt*idx0, dt*(idx1-idx0-2), None)
    
    results = {}
    results["mean"] = features.feature_list[-1].mean
    results["stdev"] = features.feature_list[-1].stdev

    return results


def analyze_long_squares(sweep_numbers, nwb_file):
    '''
    Extract features specific to long square sweeps.  From the list of sweeps, this
    will identify the rheobase sweep (lowest-amplitude stimulus inducing a spike), 
    a "hero" sweep (lowest-amplitude sweep 30-60pA above rheobase), and
    a large number of aggregate spike features for the supplied set of sweep numbers.

    Parameters
    ----------

    sweep_numbers: list
        List of long square sweep numbers.

    nwb_file: string
        File name of NWB file.
    '''

    analysis = {}

    if len(sweep_numbers) == 0:
        return analysis
    
    analysis['sweep_info'] = []
    analysis['subthresh'] = []
    analysis['fI'] = []
    features = fx.EphysFeatureExtractor()
    
    # Go through the sweeps to identify the subthreshold ones and the rheobase sweep
    rheo_amp = 1e12
    rheo_n_spikes = 1e12
    rheo_sweep_num = -1

    sweep_numbers = sorted(sweep_numbers)
    for sweep_number in sweep_numbers:

        sweep_info = {}
        sweep_info['sweep_num'] = sweep_number

        v, i, t = get_sweep_from_nwb(nwb_file, sweep_number)

        sweep_info['stim_start'], sweep_info['stim_dur'], sweep_info['stim_amp'], start_idx, end_idx = get_square_stim_characteristics(i, t)

        features.process_instance(sweep_number, v, i, t, 
                                  sweep_info['stim_start'], sweep_info['stim_dur'], 
                                  "long_square")

        n_spikes = features.feature_list[-1].mean['n_spikes']


        steady_interval_start_idx = np.nonzero(t >= sweep_info['stim_start'] + sweep_info['stim_dur'] - 0.1)[0][0]

        # if there are no spikes
        if n_spikes == 0:

            # find index of peak (absolute) response 
            if sweep_info['stim_amp'] < 0:
                peak_val_idx = np.argmin(v[start_idx:end_idx + 1]) + start_idx
            else:
                peak_val_idx = np.argmax(v[start_idx:end_idx + 1]) + start_idx

            # calculate membrane tau
            sweep_tau = calculate_membrane_tau(v, t, start_idx, peak_val_idx) * 1e3 # to ms
            
            analysis['subthresh'].append({'amp': float(sweep_info['stim_amp']), 
                                          'peak': float(v[peak_val_idx]),
                                          'peak_t': float(t[peak_val_idx]),
                                          'peak_idx': int(peak_val_idx),
                                          'steady': float(v[steady_interval_start_idx:end_idx].mean()),
                                          'tau': float(sweep_tau),
                                          'sweep_num': sweep_number,
                                          'base_vm': float(features.feature_list[-1].mean['base_v'])
                                          })
        elif sweep_info['stim_amp'] < rheo_amp:
            # there are spikes, and the stimulus amplitude 
            # is lower than the current rheobase amplitude
            rheo_n_spikes = n_spikes
            rheo_amp = sweep_info['stim_amp']
            rheo_sweep_num = sweep_number

        sweep_info['base_vm'] = float(features.feature_list[-1].mean['base_v'])

        analysis['sweep_info'].append(sweep_info)
        
        # Add point to the fI curve
        analysis['fI'].append((float(sweep_info['stim_amp']), float(n_spikes / sweep_info['stim_dur'])))
        

    if rheo_sweep_num < 0:
        logging.error("no rheo sweep found")
        return analysis

    try:
        analysis['input_resistance'] = float(calculate_input_resistance(analysis['subthresh']))
        analysis['tau'] = float(cellwide_tau(analysis['subthresh']))
    except ValueError:
        logging.error('input resistance calculation requires at least two passing, negative amplitude subthreshold long square sweeps.')
        return analysis

    analysis['v_rest'] = float(np.array([s['base_vm'] for s in analysis['sweep_info']]).mean())

    logging.info("Ri = {:.1f} MOhm".format(analysis['input_resistance']))
    logging.info("Tau_m = {:.1f} ms".format(analysis['tau']))
    logging.info("Vrest = {:.1f} mV".format(analysis['v_rest']))
    
    # Characterize the sag
    target = -100.0
    delta = 1e12
    sag_fraction = 0
    sag_vm = 0

    for s in analysis['subthresh']:
        if abs(s['peak'] - target) < delta:
            delta = abs(s['peak'] - target)
            sag_vm = s['peak']
            sag_fraction = (s['peak'] - s['steady']) / (s['peak'] - s['base_vm'])

    analysis['sag'] = float(sag_fraction)
    analysis['sag_eval_vm'] = float(sag_vm)

    # Get AP characteristics from rheobase sweep
    if rheo_sweep_num >= 0: 
        analysis['rheo_sweep_num'] = int(rheo_sweep_num)
        rheo_idx = [ i for i,sn in enumerate(sweep_numbers) if sn == rheo_sweep_num ][0]

        ft = features.feature_list[rheo_idx].mean

        v, i, t = get_sweep_from_nwb(nwb_file, rheo_sweep_num)
        
        # Average all spikes
        analysis['rheo_avg'] = {}
        analysis['rheo_avg']['up_down_ratio'] = float(ft['upstroke'] / -ft['downstroke'])
        analysis['rheo_avg']['trough'] = float(ft['f_trough'])
        analysis['rheo_avg']['fast_trough'] = float(ft['f_fast_ahp'])
        analysis['rheo_avg']['slow_trough'] = float(ft['f_slow_ahp'])
        analysis['rheo_avg']['has_true_ahp'] = bool(ft['f_trough'] < ft['base_v'])
        analysis['rheo_avg']['thresh_v'] = float(ft['threshold'])
        
        # First spike values
        analysis['rheo_spike_0'] = {}
        analysis['rheo_spike_0']['up_down_ratio'] = float(ft['spikes'][0]['upstroke'] / -ft['spikes'][0]['downstroke'])
        analysis['rheo_spike_0']['trough_v'] = float(ft['spikes'][0]['f_trough'])
        analysis['rheo_spike_0']['trough_t'] = float(ft['spikes'][0]['trough_t'] - t[start_idx])
        analysis['rheo_spike_0']['fast_trough_v'] = float(ft['spikes'][0]['f_fast_ahp'])
        analysis['rheo_spike_0']['fast_trough_t'] = float(ft['spikes'][0]['f_fast_ahp_t'] - t[start_idx])
        analysis['rheo_spike_0']['slow_trough_v'] = float(ft['spikes'][0]['f_slow_ahp'])
        analysis['rheo_spike_0']['slow_trough_t'] = float(ft['spikes'][0]['f_slow_ahp_t'] - t[start_idx])
        analysis['rheo_spike_0']['has_true_ahp'] = bool(ft['spikes'][0]['f_trough'] < ft['base_v'])
        analysis['rheo_spike_0']['thresh_v'] = float(ft['spikes'][0]['threshold'])
        analysis['rheo_spike_0']['thresh_t'] = float(ft['spikes'][0]['t'] - t[start_idx])
        analysis['rheo_spike_0']['thresh_i'] = float(i[ft['spikes'][0]['t_idx']])
        analysis['rheo_spike_0']['peak_v'] = float(ft['spikes'][0]['f_peak'])
        analysis['rheo_spike_0']['peak_t'] = float(ft['spikes'][0]['f_peak_t'] - t[start_idx])

    logging.info("Upstroke/downstroke ratio = {:.2f}".format(analysis['rheo_avg']['up_down_ratio']))
    
    # Get spike pattern characteristics from rheobase +40pA to +60pA. if none available, exit analysis.
    hero_idx = -1
    hero_range = [ 39.0, 61.0 ]

    # find hero sweeps within the range of valid pA offsets
    hero_sweeps = [ (s,i) for i,s in enumerate(analysis['sweep_info']) if (s['stim_amp'] - rheo_amp >= hero_range[0] and 
                                                                           s['stim_amp'] - rheo_amp <= hero_range[1]) ]

    if len(hero_sweeps) == 0:
        logging.error("could not find hero sweep, giving up on analysis.")
        return analysis
                
    # sort the potential hero sweeps in ascending order
    hero_sweeps.sort(key=lambda s: s[0]['stim_amp'])

    # grab the sweep with the lowest amplitude
    hero_idx = hero_sweeps[0][1]

    analysis['hero_sweep_num'] = sweep_numbers[hero_idx]
    ft = features.feature_list[hero_idx].mean

    spikes = ft['spikes']

    if 'isi_avg' in ft and ft['isi_avg'] is not None:
        avg_isi = ft['isi_avg']
        analysis['avg_isi'] = float(avg_isi)

    if 'adapt' in ft and ft['adapt'] is not None:
        analysis['adaptation_index'] = float(ft['adapt'])

    if 'latency' in ft:
        analysis['latency'] = float(ft['latency'])
    
    if len(spikes) >= 2:
        isis = [spikes[i + 1]['t'] - spikes[i]['t'] for i in range(len(spikes)-1)]

        if len(isis) > 0:
            analysis['has_delay'] = bool(ft['latency'] > avg_isi)
            analysis['delay_ratio'] = float(ft['latency'] / avg_isi)

        if len(isis) > 1:
            analysis['has_burst'] = bool(isis[0] <= 5e-3 and isis[1] <= 5e-3)
            analysis['isi_0'] = float(isis[0])
            analysis['isi_1'] = float(isis[1])

        if len(isis) > 2:
            analysis['has_pause'] = bool(check_for_pause(isis))
        
    # Figure out metric for fI curve shape - just fitting non-zero part to line
    fI_sorted = sorted(analysis['fI'], key=lambda d: d[0])
    x = np.array([d[0] for d in fI_sorted], dtype=np.float64)
    y = np.array([d[1] for d in fI_sorted], dtype=np.float64)
    last_zero_idx = np.nonzero(y)[0][0] - 1
    A = np.vstack([x[last_zero_idx:], np.ones(len(x[last_zero_idx:]))]).T
    m, c = np.linalg.lstsq(A, y[last_zero_idx:])[0]
    analysis['fI_fit_slope'] = float(m)
    
    logging.info("Linear fit to f-I curve slope = {:.2f}".format(analysis['fI_fit_slope']))

    return analysis
    

def analyze_ramps(sweep_numbers, nwb_file):
    '''
    Extract ramp-specific features from a set of sweeps in an NWB file.  These are
    primarily aggregate spike features.

    Parameters
    ----------

    sweep_numbers: list
        List of ramp sweep numbers.

    nwb_file: string
        File name of NWB file.
    '''

    analysis = {}
    analysis['sweep_info'] = []

    if len(sweep_numbers) == 0:
        return analysis
        
    features = fx.EphysFeatureExtractor()
    
    # Average values across ramps, since all ramps should have the same stimulus applied
    # and only look at the first spike
    sweep_count = 0
    value_keys = ['up_down_ratio', 'trough_v', 'trough_t', 'fast_trough_v', 'fast_trough_t', 'slow_trough_v', 'slow_trough_t', 'thresh_v', 'thresh_t', 'thresh_i', 'peak_t', 'peak_v']
    for k in value_keys:
        analysis[k] = 0
        
    for sweep_number in sorted(sweep_numbers):
        sweep_info = {}
        sweep_info['sweep_num'] = sweep_number

        v, i, t = get_sweep_from_nwb(nwb_file, sweep_number)
        
        if np.all(i == 0): # Check for weird ramp sweeps (orca conversion issue?)
            logging.warning("ramp sweep %d had all-zero stimulus.  skipping. " % sweep_number)
            continue
            
        sweep_info['stim_start'], start_idx = get_ramp_stim_characteristics(i, t)
        
        # Figure out where to end, since it goes to zero when cut off
        rev_v = v[::-1]
        nonzero_idx = np.flatnonzero(rev_v)[0]
        end_idx = len(v) - nonzero_idx - 1
            
        features.process_instance(sweep_number, v, i, t, 
                                  sweep_info['stim_start'], t[end_idx] - sweep_info['stim_start'], 
                                  'ramp')

        n_spikes = features.feature_list[-1].mean['n_spikes']
        
        
        if n_spikes > 0:
            sweep_count += 1

            spk = features.feature_list[-1].mean['spikes'][0]
            analysis['up_down_ratio'] += float(spk['upstroke'] / -spk['downstroke'])
            analysis['trough_v'] += float(spk['f_trough'])
            analysis['trough_t'] += float(spk['trough_t'] - t[start_idx])
            analysis['fast_trough_v'] += float(spk['f_fast_ahp'])
            analysis['fast_trough_t'] += float(spk['f_fast_ahp_t'] - t[start_idx])
            analysis['slow_trough_v'] += float(spk['f_slow_ahp'])
            analysis['slow_trough_t'] += float(spk['f_slow_ahp_t'] - t[start_idx])
            analysis['thresh_v'] += float(spk['threshold'])
            analysis['thresh_t'] += float(spk['t'] - t[start_idx])
            analysis['thresh_i'] += float(i[spk['t_idx']])
            analysis['peak_v'] += float(spk['f_peak'])
            analysis['peak_t'] += float(spk['f_peak_t'] - t[start_idx])

        analysis['sweep_info'].append(sweep_info)

    logging.info("Ramps with spikes: {:d}".format(sweep_count))

    if sweep_count > 0:
        for k in value_keys:
            analysis[k] /= sweep_count

        logging.info("Ramp up/down ratio: {:.1f}".format(analysis['up_down_ratio']))
        logging.info("Ramp threshold: {:.1f} mV at {:.1f} pA".format(analysis['thresh_v'], analysis['thresh_i']))
    else:
        for k in value_keys:
            analysis[k] = None

    analysis['sweep_count'] = sweep_count
    
    return analysis


def analyze_short_squares(sweep_numbers, nwb_file):
    '''
    Extract ramp-specific features from a set of sweeps in an NWB file.  Most
    of the features come from the lowest-amplitude spike-inducing sweeps.

    Parameters
    ----------

    sweep_numbers: list
        List of short square sweep numbers.

    nwb_file: string
        File name of NWB file.
    '''

    analysis = {}
    analysis['sweep_info'] = []

    if len(sweep_numbers) == 0:
        return analysis

    features = fx.EphysFeatureExtractor()
    spike_info = {}
    evoking_amps = []

    for sweep_number in sorted(sweep_numbers):
        sweep_info = {}
        sweep_info['sweep_num'] = sweep_number

        v, i, t = get_sweep_from_nwb(nwb_file, sweep_number)

        sweep_info['stim_start'], sweep_info['stim_dur'], sweep_info['stim_amp'], start_idx, end_idx = get_square_stim_characteristics(i, t)

        # We are actually looking for spikes *after* the end of the stimulus
        features.process_instance(sweep_number, v, i, t, 
                                  sweep_info['stim_start'], t[-1] - sweep_info['stim_start'],
                                  'short_square')

        n_spikes = features.feature_list[-1].mean['n_spikes']
        if n_spikes > 0:
            spike_info[sweep_number] = features.feature_list[-1].mean['spikes'][0]
            evoking_amps.append(sweep_info['stim_amp'])
        analysis['sweep_info'].append(sweep_info)
        
    # We want to analyze the smallest amplitude short_squares that have the most spikes evoked
    # So, if there is one 100pA short_square that evoked one spike, but there are 3 110pA short_squares
    # that evoke spikes, we'll analyze the 110pA ones.
    min_amp = 1e12
    min_count = 0
    amp_counts = Counter(evoking_amps)
    for amp in amp_counts:
        if amp_counts[amp] > min_count or (amp_counts[amp] == min_count and amp < min_amp):
            min_amp = amp
            min_count = amp_counts[amp]
    
    # now find the spikes at that amplitude and average their stats
    sweep_count = 0
    value_keys = ['up_down_ratio', 'trough_v', 'trough_t', 'fast_trough_v', 'fast_trough_t', 'slow_trough_v', 'slow_trough_t', 'thresh_v', 'thresh_t', 'thresh_i', 'peak_t', 'peak_v']

    for k in value_keys:
        analysis[k] = 0

    for si in analysis['sweep_info']:
        if si['stim_amp'] == min_amp and si['sweep_num'] in spike_info:
            spk = spike_info[si['sweep_num']]
            analysis['up_down_ratio'] += float(spk['upstroke'] / -spk['downstroke'])
            analysis['trough_v'] += float(spk['f_trough'])
            analysis['trough_t'] += float(spk['trough_t'] - t[start_idx])
            analysis['fast_trough_v'] += float(spk['f_fast_ahp'])
            analysis['fast_trough_t'] += float(spk['f_fast_ahp_t'] - t[start_idx])
            analysis['slow_trough_v'] += float(spk['f_slow_ahp'])
            analysis['slow_trough_t'] += float(spk['f_slow_ahp_t'] - t[start_idx])
            analysis['thresh_v'] += float(spk['threshold'])
            analysis['thresh_t'] += float(spk['t'] - si['stim_start'])
            analysis['thresh_i'] += float(si['stim_amp'])
            analysis['peak_v'] += float(spk['f_peak'])
            analysis['peak_t'] += float(spk['f_peak_t'] - t[start_idx])
            sweep_count += 1

    if sweep_count > 0:
        for k in value_keys:
            analysis[k] /= sweep_count
        analysis['repeat_amp'] = float(min_amp)
        logging.info("Short Square up/down ratio: {:.1f}".format(analysis['up_down_ratio']))
        logging.info("Short Square threshold: {:.1f} mV at {:.1f} pA".format(analysis['thresh_v'], analysis['thresh_i']))
    else:
        for k in value_keys:
            analysis[k] = None
    return analysis
    

def get_sweep_from_nwb(nwb_file, sweep_num):
    '''
    Read a sweep from an NWB file and convert Volts -> mV and Amps -> pA. 
    '''
    ds = NwbDataSet(nwb_file)
    data = ds.get_sweep(sweep_num)

    v = data['response'] * 1e3 # convert to mV
    i = data['stimulus'] * 1e12 # convert to pA

    dt = 1.0 / data['sampling_rate']
    t = np.arange(0,len(v)) * dt
    
    return (v, i, t)
        

def get_square_stim_characteristics(i, t):
    '''
    Identify the start time, duration, amplitude, start index, and
    end index of a square stimulus.  
    '''

    # Assumes that there is a test pulse followed by the stimulus square
    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)
    down_idx = np.flatnonzero(di < 0)
    
    # second square is the stimulus
    if up_idx[1] < down_idx[1]: # positive square
        start_idx = up_idx[1] + 1 # shift by one to compensate for diff()
        end_idx = down_idx[1] + 1
    else: # negative square
        start_idx = down_idx[1] + 1
        end_idx = up_idx[1] + 1
    stim_start = float(t[start_idx])
    stim_dur = float(t[end_idx] - t[start_idx])
    stim_amp = float(i[start_idx])
    return (stim_start, stim_dur, stim_amp, start_idx, end_idx)


def get_ramp_stim_characteristics(i, t):
    ''' Identify the start time and start index of a ramp sweep. '''

    # Assumes that there is a test pulse followed by the stimulus ramp
    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)
    
    start_idx = up_idx[1] + 1 # shift by one to compensate for diff()
    return (t[start_idx], start_idx)
    

def calculate_input_resistance(subthresh_data):
    ''' 
    Calculate the input resistance of a sweep using the output of a EphysFeatureExtractor instance 
    applied to long square sweeps.  This only uses the negative-going squares, and ignores
    sweeps that exceed -100 pA since some cells start to move from linear in that regime.
    '''

    i = np.array([d['amp'] for d in subthresh_data if d['amp'] < 0 and d['amp'] > -100])
    v = np.array([d['peak'] for d in subthresh_data if d['amp'] < 0 and d['amp'] > -100])
    A = np.vstack([i, np.ones(len(i))]).T
    m, c = np.linalg.lstsq(A, v)[0]
    return m * 1e3 # since MOhm = 1e3 * mV / pA


def cellwide_tau(subthresh_data):
    ''' Return the tau of the negative-going long square sweeps above -100 pA. '''

    taus = np.array([d['tau'] for d in subthresh_data if d['amp'] < 0 and d['amp'] > -100 and not np.isnan(d['tau'])])
    return taus.mean()


def calculate_membrane_tau(v, t, start_idx, peak_idx):
    ''' 
    Calculate membrane tau by fitting the response of the rheobase sweep.
    
    Parameters
    ----------

    v: np.ndarray
        array of voltages in mV

    t: np.ndarray
        array of time stamps in seconds

    start_index: int
        index of stimulus onset

    peak_idx: int
        index of the peak of the spike
    '''
    tenpct_idx, popt = fit_membrane_tau(v, t, start_idx, peak_idx)
    return 1.0 / popt[1] # seconds
    

def fit_membrane_tau(v, t, start_idx, peak_idx):
    ''' 
    Calculate membrane tau by fitting the response of the rheobase sweep.
    
    Parameters
    ----------

    v: np.ndarray
        array of voltages in mV

    t: np.ndarray
        array of time stamps in seconds

    start_index: int
        index of stimulus onset

    peak_idx: int
        index of the peak of the spike
    '''

    try:
        # fit from 10% up to peak
        if v[start_idx] > v[peak_idx]:
            tenpct_idx = np.where(v[start_idx:peak_idx] <= 0.1 * (v[peak_idx] - v[start_idx]) + v[start_idx])[0][0] + start_idx
        else:
            tenpct_idx = np.where(v[start_idx:peak_idx] >= 0.1 * (v[peak_idx] - v[start_idx]) + v[start_idx])[0][0] + start_idx

        guess = [v[tenpct_idx] - v[peak_idx], 50., v[peak_idx]]
        try:
            popt, pcov = curve_fit(exp_curve, t[tenpct_idx:peak_idx].astype(np.float64) - t[tenpct_idx], v[tenpct_idx:peak_idx].astype(np.float64), p0=guess)
        except RuntimeError:
            logging.warning("Curve fit for membrane tau failed.")
            return (np.nan, [np.nan, np.nan, np.nan])
        return (tenpct_idx, popt)
    except IndexError:
        logging.warning("Index error occurred calculating tau. Aborting calculation")
        return (np.nan, [np.nan, np.nan, np.nan])
    

def exp_curve(x, a, inv_tau, y0):
    ''' Function used for tau curve fitting '''
    return y0 + a * np.exp(-inv_tau * x)


def check_for_pause(isis):
    ''' Detect a pause given a list of inter-spike intervals. '''
    for i, isi in enumerate(isis[1:-1]):
        if isi > 3 * isis[i - 1 + 1] and isi > 3 * isis[i + 1 + 1]:
            return True
    return False


if __name__ == "__main__": pass





# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.queries.rma_api import RmaApi
from allensdk.api.queries.grid_data_api import GridDataApi
import numpy as np
import os, nrrd


class MouseConnectivityApi(RmaApi):
    '''
    HTTP Client for the Allen Mouse Brain Connectivity Atlas.
    
    See: `Mouse Connectivity API <http://help.brain-map.org/display/mouseconnectivity/API>`_
    '''
    PRODUCT_IDS = [5, 31]
    
    def __init__(self, base_uri=None):
        super(MouseConnectivityApi, self).__init__(base_uri)


    def download_annotation_volume(self, resolution, file_name):
        '''
        Download the annotation volume at a particular resolution.

        Parameters
        ----------
        
        resolution: int
            Desired resolution to download in microns.  Must be 10, 25, 50, or 100.

        file_name: string
            Where to save the annotation volume.
        '''

        try:
            os.makedirs(os.path.dirname(file_name))
        except:
            pass

        self.download_volumetric_data('annotation/ccf_2015',
                                      'annotation_%d.nrrd' % resolution,
                                      save_file_path=file_name)
        
        annotation_data, annotation_image = nrrd.read(file_name)
        
        return annotation_data, annotation_image


    def download_template_volume(self, resolution, file_name):
        '''
        Download the registration template volume at a particular resolution.

        Parameters
        ----------
        
        resolution: int
            Desired resolution to download in microns.  Must be 10, 25, 50, or 100.

        file_name: string
            Where to save the registration template volume.
        '''
        try:
            os.makedirs(os.path.dirname(file_name))
        except:
            pass

        self.download_volumetric_data('average_template',
                                      'average_template_%d.nrrd' % resolution,
                                      save_file_path=file_name)

        annotation_data, annotation_image = nrrd.read(file_name)
        
        return annotation_data, annotation_image
    
    
    def get_experiments(self,
                        structure_ids,
                        **kwargs):
        ''' 
        Fetch experiment metadata from the Mouse Brain Connectivity Atlas.
        
        Parameters
        ----------
        structure_ids : integer or list, optional
            injection structure
        
        Returns
        -------
        url : string
            The constructed URL
        '''
        criteria_list = ['[failed$eqfalse]',
                         'products[id$in%s]' % (','.join(str(i) for i in MouseConnectivityApi.PRODUCT_IDS))]
            
        if structure_ids != None:
            if type(structure_ids) is not list:
                structure_ids = [ structure_ids ]
            criteria_list.append('[id$in%s]' % ','.join(str(i) for i in structure_ids))

        criteria_string = ','.join(criteria_list)
                
        data = self.model_query('SectionDataSet',
                                criteria=criteria_string,
                                **kwargs)
        return data
    
    
    def get_manual_injection_summary(self, experiment_id):
        ''' Retrieve manual injection summary. '''

        criteria = '[id$in%d]' % (experiment_id)
        
        include = ['specimen(donor(transgenic_mouse(transgenic_lines)),',
                   'injections(structure,age)),',
                   'equalization,products']
        
        only = ['id',
                'failed',
                'storage_directory',
                'red_lower',
                'red_upper',
                'green_lower',
                'green_upper',
                'blue_lower',
                'blue_upper',
                'products.id',
                'specimen_id',
                'structure_id',
                'reference_space_id',
                'primary_injection_structure_id',
                'registration_point',
                'coordinates_ap',
                'coordinates_dv',
                'coordinates_ml',
                'angle',
                'sex',
                'strain',
                'injection_materials',
                'acronym',
                'structures.name',
                'days',
                'transgenic_mice.name',
                'transgenic_lines.name',
                'transgenic_lines.description',
                'transgenic_lines.id',
                'donors.id']
        
        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include,
                                only=only)
    
    
    def get_experiment_detail(self, experiment_id):
        '''Retrieve the experiments data.'''

        criteria = '[id$eq%d]' % (experiment_id)
        include = [ 'specimen(stereotaxic_injections(primary_injection_structure,structures,stereotaxic_injection_coordinates)),',
                    'equalization,',
                    'sub_images' ]
        order = [ "'sub_images.section_number$asc'" ]
        
        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include,
                                order=order)
    
    
    def get_projection_image_info(self,
                                  experiment_id,
                                  section_number):
        '''Fetch meta-information of one projection image.
        
        Parameters
        ----------
        experiment_id : integer
        
        section_number : integer
        
        Notes
        -----
        See: image examples under 
        `Experimental Overview and Metadata <http://help.brain-map.org/display/mouseconnectivity/API##API-ExperimentalOverviewandMetadata>`_
        for additional documentation.
        Download the image using :py:meth:`allensdk.api.queries.image_download_api.ImageDownloadApi.download_section_image`
        '''

        criteria = '[id$eq%d]' % (experiment_id)
        include = [ 'equalization,sub_images[section_number$eq%d]' % (section_number) ]

        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include)
    
    
    def build_volumetric_data_download_url(self,
                                           data,
                                           file_name,
                                           voxel_resolution=None,
                                           release=None,
                                           coordinate_framework=None):
        '''Construct url to download 3D reference model in NRRD format.
        
        Parameters
        ----------
        data : string
            'average_template', 'ara_nissl', 'annotation/ccf_2015', 'annotation/mouse_2011', or 'annotation/devmouse_2012'
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'
            
        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_ 
        for additional documentation.
        '''
        
        if voxel_resolution == None:
            voxel_resolution = 10
            
        if release == None:
            release = 'current-release'
        
        if coordinate_framework == None:
            coordinate_framework = 'mouse_ccf'
        
        url = ''.join([self.informatics_archive_endpoint,
                       '/%s/%s/' % (release, coordinate_framework),
                       data,
                       '/',
                       file_name])
        
        return url
        
    
    def download_volumetric_data(self,
                                 data,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        '''Download 3D reference model in NRRD format.
        
        Parameters
        ----------
        data : string
            'average_template', 'ara_nissl', 'annotation/ccf_2015', 'annotation/mouse_2011', or 'annotation/devmouse_2012'
        file_name : string
            
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'
            
        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_ 
        for additional documentation.
        '''
        url = self.build_volumetric_data_download_url(data,
                                                      file_name,
                                                      voxel_resolution,
                                                      release,
                                                      coordinate_framework)
        
        self.retrieve_file_over_http(url, save_file_path)
    
    
    def download_reference_aligned_image_channel_volumes(self,
                                                         data_set_id,
                                                         save_file_path=None):
        '''
        Returns
        -------
            The well known file is downloaded
        '''
        well_known_file_url = self.get_reference_aligned_image_channel_volumes_url(data_set_id)
        
        if save_file_path == None:
            save_file_path = str(data_set_id) + '.zip'
            
        self.retrieve_file_over_http(well_known_file_url, save_file_path)


    def build_reference_aligned_image_channel_volumes_url(self,
                                                          data_set_id):
        '''Construct url to download the red, green, and blue channels
        aligned to the 25um adult mouse brain reference space volume.
        
        Parameters
        ----------
        data_set_id : integer
            aka attachable_id
            
        Notes
        -----
        See: `Reference-aligned Image Channel Volumes <http://help.brain-map.org/display/mouseconnectivity/API#API-ReferencealignedImageChannelVolumes>`_ 
        for additional documentation.
        '''
        
        criteria = ['well_known_file_type',
                    "[name$eq'ImagesResampledTo25MicronARA']",
                    "[attachable_id$eq%d]" % (data_set_id)]
        
        model_stage = self.model_stage('WellKnownFile',
                                      criteria=criteria)
        
        url = self.build_query_url([model_stage])
        
        return url
    
    
    def get_reference_aligned_image_channel_volumes_url(self,
                                                        data_set_id):
        '''Retrieve the download link for a specific data set.\
        
        Notes
        -----
        See `Reference-aligned Image Channel Volumes <http://help.brain-map.org/display/mouseconnectivity/API#API-ReferencealignedImageChannelVolumes>`_
        for additional documentation.
        '''
        download_link = self.do_query(self.build_reference_aligned_image_channel_volumes_url,
                                      lambda parsed_json: str(parsed_json['msg'][0]['download_link']),
                                      data_set_id)
        
        url = self.api_url + download_link
        
        return url

    
    
    def experiment_source_search(self, **kwargs):
        '''Search over the whole projection signal statistics dataset
        to find experiments with specific projection profiles.
        
        Parameters
        ----------
        injection_structures : list of integers or strings
            Integer Structure.id or String Structure.acronym.
        target_domain : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        injection_hemisphere : string, optional
            'right' or 'left', Defaults to both hemispheres.
        target_hemisphere : string, optional
            'right' or 'left', Defaults to both hemispheres.
        transgenic_lines : list of integers or strings, optional
             Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_domain : list of integers or strings, optional
             Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.
        
        Notes
        -----
        See `Source Search <http://help.brain-map.org/display/mouseconnectivity/API#API-SourceSearch>`_,
        `Target Search <http://help.brain-map.org/display/mouseconnectivity/API#API-TargetSearch>`_,
        and 
        `service::mouse_connectivity_injection_structure <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivityinjectionstructure>`_.
        
        '''
        tuples = [ (k,v) for k,v in kwargs.iteritems() ]
        return self.service_query('mouse_connectivity_injection_structure', parameters=tuples)

    
    def experiment_spatial_search(self, **kwargs):
        '''Displays all SectionDataSets
        with projection signal density >= 0.1 at the seed point.
        This service also returns the path
        along the most dense pixels from the seed point
        to the center of each injection site..
        
        Parameters
        ----------
        seed_point : list of floats
            The coordinates of a point in 3-D SectionDataSet space.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        section_data_sets : list of integers, optional
            Ids to filter the results.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.
        
        Notes
        -----
        See `Spatial Search <http://help.brain-map.org/display/mouseconnectivity/API#API-SpatialSearch>`_
        and 
        `service::mouse_connectivity_target_spatial <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivitytargetspatial>`_.
        
        '''

        tuples = [ (k,v) for k,v in kwargs.iteritems() ]
        return self.service_query('mouse_connectivity_target_spatial', parameters=tuples)
    
    
    def experiment_injection_coordinate_search(self, **kwargs):
        '''User specifies a seed location within the 3D reference space.
        The service returns a rank list of experiments
        by distance of its injection site to the specified seed location.
        
        Parameters
        ----------
        seed_point : list of floats
            The coordinates of a point in 3-D SectionDataSet space.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.
        
        Notes
        -----
        See `Injection Coordinate Search <http://help.brain-map.org/display/mouseconnectivity/API#API-InjectionCoordinateSearch>`_
        and 
        `service::mouse_connectivity_injection_coordinate <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivityinjectioncoordinate>`_.
        
        '''
        tuples = [ (k,v) for k,v in kwargs.iteritems() ]
        return self.service_query('mouse_connectivity_injection_coordinate', parameters=tuples)
    
    
    def experiment_correlation_search(self, **kwargs):
        '''Select a seed experiment and a domain over
        which the similarity comparison is to be made.
        
        
        Parameters
        ----------
        row : integer
            SectionDataSet.id to correlate against.
        structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        hemisphere : string, optional
            Use 'right' or 'left'. Defaults to both hemispheres.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.
        
        Notes
        -----
        See `Correlation Search <http://help.brain-map.org/display/mouseconnectivity/API#API-CorrelationSearch>`_
        and 
        `service::mouse_connectivity_correlation <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivitycorrelation>`_.
        
        '''
        tuples = [ (k,v) for k,v in kwargs.iteritems() ]
        return self.service_query('mouse_connectivity_correlation', parameters=tuples)


    def get_structure_unionizes(self,
                                experiment_ids,
                                is_injection=None,
                                structure_name=None,
                                structure_ids=None,
                                hemisphere_ids=None,
                                normalized_projection_volume_limit=None,
                                include=None,
                                debug=None,
                                order=None):

        experiment_filter = '[section_data_set_id$in%s]' %\
                            ','.join(str(i) for i in experiment_ids)
        
        if is_injection == True:
            is_injection_filter = '[is_injection$eqtrue]'
        elif is_injection == False:
            is_injection_filter = '[is_injection$eqfalse]'
        else:
            is_injection_filter = ''
        
        if normalized_projection_volume_limit != None:
            volume_filter = '[normalized_projection_volume$gt%f]' %\
                            (normalized_projection_volume_limit)
        else:
            volume_filter = ''

        
        if hemisphere_ids is not None:
            hemisphere_filter = '[hemisphere_id$in%s]' %\
                ','.join(str(h) for h in hemisphere_ids)
        else:
            hemisphere_filter = ''
        
        if structure_name != None:
            structure_filter = ",structure[name$eq'%s']" % (structure_name)
        elif structure_ids != None:
            structure_filter = '[structure_id$in%s]' %\
                               ','.join(str(i) for i in structure_ids)
        else:
            structure_filter = ''
        
        return self.model_query(
                'ProjectionStructureUnionize',
                criteria=''.join([experiment_filter,
                                  is_injection_filter,
                                  volume_filter,
                                  hemisphere_filter,
                                  structure_filter]),
                include=include,
                order=order,
                num_rows='all',
                debug=debug,
                count=False)

    def download_injection_density(self, path, experiment_id, resolution):
        return GridDataApi().download_projection_grid_data(
            experiment_id, [ GridDataApi.INJECTION_DENSITY ], resolution, path)


    def download_projection_density(self, path, experiment_id, resolution):
        return GridDataApi().download_projection_grid_data(
            experiment_id, [ GridDataApi.PROJECTION_DENSITY ], resolution, path)


    def download_injection_fraction(self, path, experiment_id, resolution):
        return GridDataApi().download_projection_grid_data(
            experiment_id, [ GridDataApi.INJECTION_FRACTION ], resolution, path)


    def download_data_mask(self, path, experiment_id, resolution):
        return GridDataApi().download_projection_grid_data(
           experiment_id, [ GridDataApi.DATA_MASK ], resolution, path)


    def calculate_injection_centroid(self,
                                     injection_density,
                                     injection_fraction,
                                     resolution=25):
        '''
        Compute the centroid of an injection site.
        
        Parameters
        ----------
        
        injection_density: np.ndarray
            The injection density volume of an experiment

        injection_fraction: np.ndarray
            The injection fraction volume of an experiment

        '''

        # find all voxels with injection_fraction > 0
        injection_voxels = np.nonzero(injection_fraction)
        injection_density_computed = np.multiply(injection_density[injection_voxels],
                                                 injection_fraction[injection_voxels]) 
        sum_density = np.sum(injection_density_computed)
    
        # compute centroid in CCF coordinates
        if sum_density > 0 :
            centroid = np.dot(injection_density_computed,
                              zip(*injection_voxels)) / sum_density * resolution
        else:
            centroid = None
        
        return centroid

